import { AddNewInvite, ReinstateInvite, ValidateCaptcha } from "./auth";
import db from "external/mongo/db";
import t from "tap";
import { MockJSONFetch } from "test-utils/mock-fetch";
import ResetDBState from "test-utils/resets";

t.test("#ReinstateInvite", (t) => {
	t.beforeEach(ResetDBState);

	t.test("Should change the 'consumed' property of an invite to true.", async (t) => {
		// mock insert
		const inviteDoc = await db.invites.insert({
			code: "foobar",
			consumed: true,
			createdBy: 1,
			createdAt: 1,
			consumedAt: 2,
			consumedBy: 2,
		});

		const response = await ReinstateInvite(inviteDoc.code);

		t.equal(response.nModified, 1, "Should modify one document");

		const invite2 = await db.invites.findOne({
			// lol
			code: inviteDoc.code,
		});

		t.equal(invite2!.consumed, false, "Should no longer be consumed");
		t.equal(invite2!.consumedAt, null, "Should revoke when it was consumed.");
		t.equal(invite2!.consumedBy, null, "Should revoke who it was consumed by.");

		t.end();
	});

	t.end();
});

t.test("#AddNewInvite", (t) => {
	t.beforeEach(ResetDBState);

	t.test("Should create a new invite from a given user", async (t) => {
		const userDoc = await db.users.findOne({ id: 1 });

		const result = await AddNewInvite(userDoc!);

		t.equal(result.createdBy, userDoc!.id, "Invite should be created by the requesting user.");
		t.equal(result.consumed, false, "Invite should not be consumed.");

		// was created +/- 6 seconds from now. This is perhaps too lenient, but we're only really testing its just around now ish.
		t.ok(Math.abs(result.createdAt - Date.now()) <= 6000, "Invite was created roughly now.");

		t.match(result.code, /^[0-9a-f]{40}$/u, "Invite code should be a 40 character hex string.");
	});

	t.end();
});

t.test("#ValidateCaptcha", async (t) => {
	t.equal(
		await ValidateCaptcha(
			"200",
			"bar",
			MockJSONFetch({
				"https://www.google.com/recaptcha/api/siteverify?secret=unused&response=200&remoteip=bar":
					{
						success: true,
					},
			})
		),
		true,
		"Validates captcha when sucess return is true"
	);

	t.equal(
		await ValidateCaptcha(
			"400",
			"bar",
			MockJSONFetch({
				"https://www.google.com/recaptcha/api/siteverify?secret=unused&response=400&remoteip=bar":
					{
						success: false,
					},
			})
		),
		false,
		"Invalidates captcha when success return is not true"
	);

	t.end();
});
